@@ -37,6 +37,8 @@
 #include "ecma119.h"
 #include "eltorito.h"
 #include "system_area.h"
+#include "image.h"
+#include "messages.h"
 
 
 /* This code stems from syslinux-3.72/utils/isohybrid, a perl script
@@ -60,7 +62,7 @@ license from above stem licenses, typically from LGPL.
 In case its generosity is needed, here is the 2-clause BSD license:
 
 make_isohybrid_mbr.c is copyright 2002-2008 H. Peter Anvin
-                              and 2008-2014 Thomas Schmitt
+                              and 2008-2015 Thomas Schmitt
 
 1. Redistributions of source code must retain the above copyright notice,
    this list of conditions and the following disclaimer.
@@ -156,7 +158,6 @@ int make_isohybrid_mbr(int bin_lba, int *img_blocks, char *mbr, int flag)
 
     /* For generating a weak random number */
     struct timeval tv;
-    struct timezone tz;
 
     if (bin_lba < 0 || bin_lba >= (1 << 29))
         return (0); /* 1 TB limit of signed 32 bit addressing of 512 byte blocks */
@@ -211,7 +212,7 @@ int make_isohybrid_mbr(int bin_lba, int *img_blocks, char *mbr, int flag)
      from. An environment variable ?
      125: Whatever, i use some 32-bit random value with no crypto strength.
      */
-    gettimeofday(&tv, &tz);
+    gettimeofday(&tv, NULL);
     id = 0xffffffff & (tv.tv_sec ^ (tv.tv_usec * 2000));
 
     /*
@@ -400,12 +401,15 @@ int lba512chs_to_buf(char **wpt, off_t lba, int head_count, int sector_count)
 
 /* Find out whether GPT and APM are desired
    flag bit0 = register APM and GPT requests in Ecma119Image
+        bit1 = do not asses and register APM
+        bit2 = do not register overall GPT partition
 */
 int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
                              int *apm_count, int flag)
 {
     int i, ilx_opts, j, ret, num_img;
     uint32_t block_count;
+    uint64_t start_block;
     uint8_t gpt_name[72];
     static uint8_t zero_uuid[16] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};
     static uint8_t basic_data_uuid[16] = {
@@ -428,11 +432,13 @@ int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
         num_img = 0;
     for (i = 0; i < num_img; i++) {
         ilx_opts = t->catalog->bootimages[i]->isolinux_options;
-        if (((ilx_opts >> 2) & 63) == 1 || ((ilx_opts >> 2) & 63) == 2) {
+        if ((((ilx_opts >> 2) & 63) == 1 || ((ilx_opts >> 2) & 63) == 2) &&
+           !(t->boot_appended_idx[i] >= 0 && t->opts->appended_as_gpt)) {
             if (*gpt_count < 128)
-                gpt_idx[*gpt_count]= i;
+                gpt_idx[*gpt_count] = i;
             (*gpt_count)++;
-            if ((flag & 1) && t->bootsrc[i] != NULL) {
+            if ((flag & 1) &&
+                (t->bootsrc[i] != NULL || t->boot_appended_idx[i] >= 0)) {
                 /* Register GPT entry */
                 memset(gpt_name, 0, 72);
                 sprintf((char *) gpt_name, "ISOHybrid%d", *gpt_count);
@@ -441,27 +447,44 @@ int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
                     uuid = hfs_uuid;
                 else
                     uuid = basic_data_uuid;
-                block_count = 0;
-                for (j = 0; j < t->bootsrc[i]->nsections; j++)
-                     block_count += t->bootsrc[i]->sections[j].size / 2048;
+                if (t->boot_appended_idx[i] >= 0) {
+                    block_count = t->appended_part_size[
+                                                      t->boot_appended_idx[i]];
+                    start_block = ((uint64_t) t->appended_part_start[
+                                                 t->boot_appended_idx[i]]) * 4;
+                } else {
+                    block_count = 0;
+                    for (j = 0; j < t->bootsrc[i]->nsections; j++)
+                        block_count += t->bootsrc[i]->sections[j].size / 2048;
+                    start_block = ((uint64_t) t->bootsrc[i]->sections[0].block)
+                                  * 4;
+                }
                 ret = iso_quick_gpt_entry(
                             t->gpt_req, &(t->gpt_req_count),
-                            ((uint64_t) t->bootsrc[i]->sections[0].block) * 4,
-                            ((uint64_t) block_count) * 4,
+                            start_block, ((uint64_t) block_count) * 4,
                             uuid, zero_uuid, gpt_flags, (uint8_t *) gpt_name);
                 if (ret < 0)
                     return ret;
             }
         }
-        if (ilx_opts & 256) {
+        if ((ilx_opts & 256) && !(flag & 2)) {
             (*apm_count)++;
-            if ((flag & 1) && t->bootsrc[i] != NULL) {
+            if ((flag & 1) &&
+                (t->bootsrc[i] != NULL || t->boot_appended_idx[i] >= 0)) {
                 /* Register APM entry */
-                block_count = 0;
-                for (j = 0; j < t->bootsrc[i]->nsections; j++)
-                     block_count += t->bootsrc[i]->sections[j].size / 2048;
+                if (t->boot_appended_idx[i] >= 0) {
+                    block_count = t->appended_part_size[
+                                                      t->boot_appended_idx[i]];
+                    start_block = t->appended_part_start[
+                                                      t->boot_appended_idx[i]];
+                } else {
+                    block_count = 0;
+                    for (j = 0; j < t->bootsrc[i]->nsections; j++)
+                        block_count += t->bootsrc[i]->sections[j].size / 2048;
+                    start_block = t->bootsrc[i]->sections[0].block;
+                }
                 ret = iso_quick_apm_entry(t->apm_req, &(t->apm_req_count),
-                                          t->bootsrc[i]->sections[0].block,
+                                          (uint32_t) start_block,
                                           block_count, "EFI", "Apple_HFS");
                 if (ret < 0)
                     return ret;
@@ -471,7 +494,12 @@ int assess_isohybrid_gpt_apm(Ecma119Image *t, int *gpt_count, int gpt_idx[128],
             }
         }
     }
-    if ((flag & 1) && *gpt_count > 0) {
+    if (*gpt_count > 0 && !(flag & 4)) {
+        (*gpt_count)++;
+        if (*gpt_count < 128)
+            gpt_idx[*gpt_count] = -1;
+    }
+    if ((flag & 1) && *gpt_count > 0 && !(flag & 4)) {
         /* Register overall GPT partition */
         memset(gpt_name, 0, 72);
         sprintf((char *) gpt_name, "ISOHybrid");
@@ -536,19 +564,28 @@ static int insert_apm_head(uint8_t *buf, int apm_count)
 static int gpt_images_as_mbr_partitions(Ecma119Image *t, char *wpt,
                                         int gpt_idx[128], int *gpt_cursor)
 {
-    int ilx_opts;
+    int ilx_opts, skip = 0;
     off_t hd_blocks;
     static uint8_t dummy_chs[3] = {
         0xfe, 0xff, 0xff, 
     };
 
+    if (gpt_idx[*gpt_cursor] < 0)
+        skip = 1;
+    else if (t->bootsrc[gpt_idx[*gpt_cursor]] == NULL)
+        skip = 1;
+    if (skip) {
+        (*gpt_cursor)++;
+        return 2;
+    }
+
     wpt[0] = 0;
     memcpy(wpt + 1, dummy_chs, 3);
     ilx_opts = t->catalog->bootimages[gpt_idx[*gpt_cursor]]->isolinux_options;
     if (((ilx_opts >> 2) & 63) == 2)
         wpt[4] = 0x00; /* HFS gets marked as "Empty" */
     else
-        ((unsigned char *) wpt)[4] = 0xef; /* "EFI (FAT-12/16/" */
+        ((unsigned char *) wpt)[4] = 0xef; /* "EFI (FAT-12/16)" */
 
     memcpy(wpt + 5, dummy_chs, 3);
 
@@ -567,8 +604,28 @@ static int gpt_images_as_mbr_partitions(Ecma119Image *t, char *wpt,
 }
 
 
+/* For generating a weak random number */
+static uint32_t iso_make_mbr_id(Ecma119Image *t, int flag)
+{
+    uint32_t id;
+    struct timeval tv;
+
+    if(t->opts->vol_uuid[0]) {
+        id = iso_crc32_gpt((unsigned char *) t->opts->vol_uuid, 16, 0);
+    } else if(t->opts->vol_modification_time > 0) {
+        id = iso_crc32_gpt((unsigned char *) &(t->opts->vol_modification_time),
+                           sizeof(time_t), 0);
+    } else {
+        gettimeofday(&tv, NULL);
+        id = 0xffffffff & (tv.tv_sec ^ (tv.tv_usec * 2000));
+    }
+    return id;
+}
+
+
 /*
  * @param flag  bit0= make own random MBR Id from current time
+ *                    >>> or from overridden modification time
  *              bit1= create protective MBR as of UEFI/GPT specs
  */
 int make_isolinux_mbr(uint32_t *img_blocks, Ecma119Image *t,
@@ -580,10 +637,14 @@ int make_isolinux_mbr(uint32_t *img_blocks, Ecma119Image *t,
     char *wpt;
     uint32_t boot_lba;
     int head_count, sector_count, ret;
-    int gpt_count = 0, gpt_idx[128], apm_count = 0, gpt_cursor;
-    /* For generating a weak random number */
-    struct timeval tv;
-    struct timezone tz;
+    int gpt_count = 0, gpt_idx[128], apm_count = 0, gpt_cursor, i;
+
+    if (t->bootsrc[0] == NULL)
+        return iso_msg_submit(t->image->id, ISO_BOOT_IMAGE_NOT_VALID, 0,
+      "Cannot refer by isohybrid MBR to data outside of ISO 9660 filesystem.");
+
+    for (i = 0; i < 128; i++)
+        gpt_idx[i] = -1;
 
     if (flag & 2) {
         part_number = 1;
@@ -625,8 +686,7 @@ int make_isolinux_mbr(uint32_t *img_blocks, Ecma119Image *t,
        (here some 32-bit random value with no crypto strength)
     */
     if (flag & 1) {
-        gettimeofday(&tv, &tz);
-        id = 0xffffffff & (tv.tv_sec ^ (tv.tv_usec * 2000));
+        id = iso_make_mbr_id(t, 0);
         lsb_to_buf(&wpt, id, 32, 0);
     } else {
         wpt+= 4;
